from speechcraft import text2voice, voice2voice, voice2embedding
from media_toolkit import AudioFile

sample_text = "I love society [laughs]! [happy] What a day to make voice overs with artificial intelligence."

output_dir = "test_files/output"

# test text2speech
audio_numpy, sample_rate = text2voice(sample_text)
audio = AudioFile().from_np_array(audio_numpy, sr=sample_rate)
audio.save(f"{output_dir}/en_speaker_3_i_love_socaity.wav")
#
## test voice cloning
embedding = voice2embedding(audio_file="test_files/voice_clone_test_voice_1.wav", voice_name="hermine").save_to_speaker_lib()
tts_new_speaker, sample_rate = text2voice(sample_text, voice="hermine")
audio_with_cloned_voice = AudioFile().from_np_array(tts_new_speaker, sr=sample_rate)
audio_with_cloned_voice.save(f"{output_dir}/hermine_i_love_socaity.wav")

# test voice2voice
v2v_audio_np, sample_rate = voice2voice(audio_file="test_files/voice_clone_test_voice_2.wav", voice_name="hermine")
v2v_audio = AudioFile().from_np_array(v2v_audio_np, sr=sample_rate)
v2v_audio.save(f"{output_dir}/potter_to_hermine.wav")




import os
import uuid
from io import BytesIO
from typing import Union

import numpy as np

from speechcraft.settings import EMBEDDINGS_DIR


class VoiceEmbedding:
    def __init__(
            self,
            codes: np.array,
            semantic_tokens: np.array,
            name: str = "new_speaker"
    ):
        self.name = name
        self.fine_prompt = codes
        self.coarse_prompt = codes[:2, :]
        self.semantic_prompt = semantic_tokens

    def save_to_speaker_lib(self):
        self.save(EMBEDDINGS_DIR)
        return self

    def save(self, folder: str):
        speaker_embedding_file = os.path.join(folder, f"{self.name}.npz")
        np.savez(speaker_embedding_file,
                 fine_prompt=self.fine_prompt,
                 coarse_prompt=self.coarse_prompt,
                 semantic_prompt=self.semantic_prompt
        )
        return speaker_embedding_file

    def to_bytes_io(self):
        f = BytesIO()
        np.savez(
           f, fine_prompt=self.fine_prompt, coarse_prompt=self.coarse_prompt,semantic_prompt=self.semantic_prompt
        )
        f.seek(0)
        return f

    @staticmethod
    def load(path_or_speaker_name: str):
        """
        Load a VoiceEmbedding from a file path or speaker nam.
        """
        if not os.path.exists(path_or_speaker_name):
            # add .npz extension
            if not path_or_speaker_name.endswith(".npz"):
                path_or_speaker_name = path_or_speaker_name + ".npz"

            # add default speaker dir
            path_or_speaker_name = os.path.join(EMBEDDINGS_DIR, path_or_speaker_name)


        if not os.path.exists(path_or_speaker_name):
            raise FileNotFoundError(f"Speaker embedding {path_or_speaker_name} not found")

        # Load data
        with np.load(path_or_speaker_name) as data:
            codes = data['fine_prompt']
            semantic_tokens = data['semantic_prompt']

        # return new VoiceEmbedding
        speaker_name = path_or_speaker_name.split("/")[-1].split(".")[0]
        return VoiceEmbedding(name=speaker_name, codes=codes, semantic_tokens=semantic_tokens)

    @staticmethod
    def load_from_bytes_io(bytes_io: BytesIO, speaker_name: str = None):
        with np.load(bytes_io) as data:
            codes = data['fine_prompt']
            semantic_tokens = data['semantic_prompt']

        if speaker_name is None or len(speaker_name) == 0:
            speaker_name = f"embedding_{uuid.uuid4()}"

        return VoiceEmbedding(name=speaker_name, codes=codes, semantic_tokens=semantic_tokens)

    @staticmethod
    def load_from_speaker_lib(speaker_name: str):
        return VoiceEmbedding.load(os.path.join(EMBEDDINGS_DIR, f"{speaker_name}.npz"))

    def __getitem__(self, key):
        """
        Support of ["semantic_tokens"] and ["fine_prompt"] and ["coarse_prompt"] syntax.
        :param key: str
        """
        if hasattr(self, key):
            return getattr(self, key)
        raise KeyError(f"Key {key} not found in VoiceEmbedding")

    def __setitem__(self, key, value):
        """
        Support of ["semantic_tokens"] and ["fine_prompt"] and ["coarse_prompt"] syntax.
        :param key: str
        """
        if hasattr(self, key):
            setattr(self, key, value)
        else:
            raise KeyError(f"Key {key} not found in VoiceEmbedding")



from typing import Dict, Optional, Union

import numpy as np

from .generation import codec_decode, generate_coarse, generate_fine, generate_text_semantic


def text_to_semantic(
    text: str,
    history_prompt: Optional[Union[Dict, str]] = None,
    temp: float = 0.7,
):
    """Generate semantic array from text.

    Args:
        text: text to be turned into audio
        history_prompt: history choice for audio cloning
        temp: generation temperature (1.0 more diverse, 0.0 more conservative)

    Returns:
        numpy semantic array to be fed into `semantic_to_waveform`
    """
    x_semantic = generate_text_semantic(
        text,
        history_prompt=history_prompt,
        temp=temp,
        use_kv_caching=True
    )
    return x_semantic


def semantic_to_waveform(
    semantic_tokens: np.ndarray,
    history_prompt: Optional[Union[Dict, str]] = None,
    temp: float = 0.7,
    max_coarse_history: int = 300,
    output_full: bool = False,
    progress_update_func: callable = None
):
    """Generate audio array from semantic input.

    Args:
        semantic_tokens: semantic token output from `text_to_semantic`
        history_prompt: history choice for audio cloning
        temp: generation temperature (1.0 more diverse, 0.0 more conservative)
        max_coarse_history: history influence. Min 60 (faster), max 630 (more context)
        progress_update_func: a callable to update the progress of the task. Called with progress float in [0, 1]
        output_full: return full generation to be used as a history prompt

    Returns:
        numpy audio array at sample frequency 24khz
    """
    coarse_tokens = generate_coarse(
        semantic_tokens,
        history_prompt=history_prompt,
        temp=temp,
        max_coarse_history=max_coarse_history,
        use_kv_caching=True,
        progress_update_func=progress_update_func
    )
    fine_tokens = generate_fine(
        coarse_tokens,
        history_prompt=history_prompt,
        temp=0.5,
    )
    audio_arr = codec_decode(fine_tokens)
    if output_full:
        full_generation = {
            "semantic_prompt": semantic_tokens,
            "coarse_prompt": coarse_tokens,
            "fine_prompt": fine_tokens,
        }
        return full_generation, audio_arr
    return audio_arr


def save_as_prompt(filepath, full_generation):
    assert(filepath.endswith(".npz"))
    assert(isinstance(full_generation, dict))
    assert("semantic_prompt" in full_generation)
    assert("coarse_prompt" in full_generation)
    assert("fine_prompt" in full_generation)
    np.savez(filepath, **full_generation)


def generate_audio(
    text: str,
    history_prompt: Optional[Union[Dict, str]] = None,
    text_temp: float = 0.7,
    waveform_temp: float = 0.7,
    output_full: bool = False,
    progress_update_func: callable = None
):
    """Generate audio array from input text.

    Args:
        text: text to be turned into audio
        history_prompt: history choice for audio cloning
        text_temp: generation temperature (1.0 more diverse, 0.0 more conservative)
        waveform_temp: generation temperature (1.0 more diverse, 0.0 more conservative)
        output_full: return full generation to be used as a history prompt
        progress_update_func: a callable to update the progress of the task.
            Called like progress_update_function(x) with x in [0, 1]
    Returns:
        numpy audio array at sample frequency 24khz
    """
    semantic_tokens = text_to_semantic(
        text,
        history_prompt=history_prompt,
        temp=text_temp
    )
    out = semantic_to_waveform(
        semantic_tokens,
        history_prompt=history_prompt,
        temp=waveform_temp,
        output_full=output_full,
        progress_update_func=progress_update_func
    )
    if output_full:
        full_generation, audio_arr = out
        return full_generation, audio_arr
    else:
        audio_arr = out
    return audio_arr



import os.path

from speechcraft.core.generation import preload_models, load_codec_model
from speechcraft.settings import MODELS_DIR, USE_GPU
from speechcraft.supp.utils import get_cpu_or_gpu

def get_hubert_manager_and_model(install_path: str = None):
    # Imports need to be here to avoid "circular" import error.
    # The reason in particular is not as clear. But works this way.
    from speechcraft.voice_cloning.custom_tokenizer import CustomTokenizer
    from speechcraft.voice_cloning.hubert_manager import HuBERTManager
    from speechcraft.voice_cloning.pre_kmeans_hubert import CustomHubert

    if install_path is None:
        install_path = MODELS_DIR

    # large huber pair
    huber_model_name = 'hubert_base_ls960_23.pth'
    hubert_model_path = os.path.join(install_path, huber_model_name)

    # small huber pair
    # huber_model_name = 'quantifier_V1_hubert_base_ls960_14.pth'

    # f"{tokenizer_lang}_tokenizer.pth"
    tokenizer_model_path = os.path.join(install_path, f"tokenizer_{huber_model_name}")

    hubert_manager = HuBERTManager()
    hubert_manager.make_sure_hubert_installed(model_path=hubert_model_path)
    hubert_manager.make_sure_tokenizer_installed(model=huber_model_name,
                                                 local_tokenizer_path=tokenizer_model_path)

    device = get_cpu_or_gpu()

    hubert_model = CustomHubert(checkpoint_path=hubert_model_path).to(device)
    meta_encodec_model = load_codec_model(use_gpu=True)

    # Load the CustomTokenizer model
    tokenizer = CustomTokenizer.load_from_checkpoint(tokenizer_model_path, map_location=device)

    return hubert_manager, hubert_model, meta_encodec_model, tokenizer


def make_sure_models_are_downloaded(install_path: str = None):
    # From https://github.com/gitmylo/bark-voice-cloning-HuBERT-quantizer

    # download and load all models
    preload_models(
        text_use_gpu=USE_GPU,
        text_use_small=False,
        coarse_use_gpu=USE_GPU,
        coarse_use_small=False,
        fine_use_gpu=USE_GPU,
        fine_use_small=False,
        codec_use_gpu=USE_GPU,
        force_reload=False,
        path=install_path
    )


def download_all_models_init(install_path: str = None):
    if install_path is None:
        install_path = MODELS_DIR

    # create models folder if not exists
    if not os.path.isdir(install_path):
        os.makedirs(install_path, exist_ok=True)

    # files for voice cloning
    if not os.path.isfile(os.path.join(install_path, 'hubert_base_ls960_23.pth')):
        get_hubert_manager_and_model(install_path=install_path)

    # download files for speechcraft
    if not os.path.isfile(os.path.join(install_path, 'coarse_2.pt')):
        make_sure_models_are_downloaded(install_path=install_path)





import argparse
import os

from scipy.io.wavfile import write as write_wav
from speechcraft.core.api import generate_audio
from speechcraft.core.generation import SAMPLE_RATE


def cli():
    """Commandline interface."""
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--text", type=str, help="text to be turned into audio")
    parser.add_argument(
        "--output_filename",
        type=str,
        default="bark_generation.wav",
        help="output audio file name",
    )
    parser.add_argument("--output_dir", type=str, default=".", help="directory to save the outputs")
    parser.add_argument(
        "--history_prompt",
        type=str,
        default=None,
        help="history choice for audio cloning, be path to the .npz file.",
    )
    parser.add_argument(
        "--text_temp",
        default=0.7,
        type=float,
        help="generation temperature (1.0 more diverse, 0.0 more conservative)",
    )
    parser.add_argument(
        "--waveform_temp",
        default=0.7,
        type=float,
        help="generation temperature (1.0 more diverse, 0.0 more conservative)",
    )
    parser.add_argument("--silent", default=False, type=bool, help="disable progress bar")
    parser.add_argument(
        "--output_full",
        default=False,
        type=bool,
        help="return full generation to be used as a history prompt",
    )

    args = vars(parser.parse_args())
    input_text: str = args.get("text")
    output_filename: str = args.get("output_filename")
    output_dir: str = args.get("output_dir")
    history_prompt: str = args.get("history_prompt")
    text_temp: float = args.get("text_temp")
    waveform_temp: float = args.get("waveform_temp")
    output_full: bool = args.get("output_full")

    try:
        os.makedirs(output_dir, exist_ok=True)
        generated_audio = generate_audio(
            input_text,
            history_prompt=history_prompt,
            text_temp=text_temp,
            waveform_temp=waveform_temp,
            output_full=output_full,
        )
        output_file_path = os.path.join(output_dir, output_filename)
        write_wav(output_file_path, SAMPLE_RATE, generated_audio)
        print(f"Done! Output audio file is saved at: '{output_file_path}'")
    except Exception as e:
        print(f"Oops, an error occurred: {e}")



from typing import Union

from speechcraft.core.voice_embedding import VoiceEmbedding
from speechcraft.core.generation import SAMPLE_RATE, codec_decode, generate_coarse, generate_fine, generate_text_semantic
from speechcraft.supp.model_downloader import make_sure_models_are_downloaded
from speechcraft.settings import MODELS_DIR


def text2voice_advanced(
        text_prompt: str,
        speaker_name: str = None,
        semantic_temp=0.7,
        semantic_top_k=50,
        semantic_top_p=0.95,
        coarse_temp=0.7,
        coarse_top_k=50,
        coarse_top_p=0.95,
        fine_temp=0.5,
        use_semantic_history_prompt=True,
        use_coarse_history_prompt=True,
        use_fine_history_prompt=True,
        output_full=False,
        progress_update_func: callable = None
):
    """
    :param text_prompt:
    :param semantic_temp:
    :param semantic_top_k:
    :param semantic_top_p:
    :param coarse_temp:
    :param coarse_top_k:
    :param coarse_top_p:
    :param fine_temp:
    :param speaker_name:
    :param use_semantic_history_prompt:
    :param use_coarse_history_prompt:
    :param use_fine_history_prompt:
    :param output_full:
    :param progress_update_func: a callable to update the progress of the task.
    :return: audio array and sample rate
    """

    # generation with more control

    # each generate block is 30% of progress thus we need to modify progress function
    total_progress = 0.0
    if progress_update_func is not None:
        def progress_update_func_block(x):
            nonlocal total_progress
            prev_progress = total_progress
            curr_progress = total_progress + round(x * 0.3, 2)
            if prev_progress != curr_progress:
                total_progress = curr_progress
                progress_update_func(total_progress + round(x * 0.3, 2))
    else:
        progress_update_func_block = None  # no progress displayed or calculated

    x_semantic = generate_text_semantic(
        text_prompt,
        history_prompt=speaker_name if use_semantic_history_prompt else None,
        temp=semantic_temp,
        top_k=semantic_top_k,
        top_p=semantic_top_p,
        progress_update_func=progress_update_func_block
    )

    total_progress = 0.3

    x_coarse_gen = generate_coarse(
        x_semantic,
        history_prompt=speaker_name if use_coarse_history_prompt else None,
        temp=coarse_temp,
        top_k=coarse_top_k,
        top_p=coarse_top_p,
        progress_update_func=progress_update_func_block
    )

    total_progress = 0.6

    x_fine_gen = generate_fine(
        x_coarse_gen,
        history_prompt=speaker_name if use_fine_history_prompt else None,
        temp=fine_temp,
        progress_update_func=progress_update_func_block
    )

    if output_full:
        full_generation = {
            'semantic_prompt': x_semantic,
            'coarse_prompt': x_coarse_gen,
            'fine_prompt': x_fine_gen,
        }
        return full_generation, codec_decode(x_fine_gen)
    return codec_decode(x_fine_gen)


def text2voice(
        text: str,
        voice: Union[str, VoiceEmbedding] = "en_speaker_3",
        semantic_temp=0.7,
        semantic_top_k=50,
        semantic_top_p=0.95,
        coarse_temp=0.7,
        coarse_top_k=50,
        coarse_top_p=0.95,
        fine_temp=0.5,
        progress_update_func: callable = None
) -> tuple:
    """
    :param text:
    :param voice: speaker name, path to the embedding or the VoiceEmbedding itself.
    :param semantic_temp:
    :param semantic_top_k:
    :param semantic_top_p:
    :param coarse_temp:
    :param coarse_top_k:
    :param coarse_top_p:
    :param fine_temp:
     :param progress_update_func: a callable to update the progress of the task.
        Called like progress_update_function(x) with x in [0, 1]
    :return: audio array and sample rate
    """

    make_sure_models_are_downloaded(install_path=MODELS_DIR)

    full_generation, audio_array = text2voice_advanced(
        text,
        semantic_temp=semantic_temp,
        semantic_top_k=semantic_top_k,
        semantic_top_p=semantic_top_p,
        coarse_temp=coarse_temp,
        coarse_top_k=coarse_top_k,
        coarse_top_p=coarse_top_p,
        fine_temp=fine_temp,
        speaker_name=voice,
        use_semantic_history_prompt=True,
        use_coarse_history_prompt=True,
        use_fine_history_prompt=True,
        output_full=True,
        progress_update_func=progress_update_func
    )
    return audio_array, SAMPLE_RATE



"""
Modified HuBERT model without kmeans.
Original author: https://github.com/lucidrains/
Modified by: https://www.github.com/gitmylo/
License: MIT
"""

# Modified code from https://github.com/lucidrains/audiolm-pytorch/blob/main/audiolm_pytorch/hubert_kmeans.py

from pathlib import Path

import fairseq
import torch
from torch import nn
from einops import pack, unpack

# import fairseq

from torchaudio.functional import resample

import logging
logging.root.setLevel(logging.ERROR)


def exists(val):
    return val is not None


def default(val, d):
    return val if exists(val) else d


class CustomHubert(nn.Module):
    """
    checkpoint and kmeans can be downloaded at https://github.com/facebookresearch/fairseq/tree/main/examples/hubert
    or you can train your own
    """

    def __init__(
        self,
        checkpoint_path,
        target_sample_hz=16000,
        seq_len_multiple_of=None,
        output_layer=9,
        device=None
    ):
        super().__init__()
        self.target_sample_hz = target_sample_hz
        self.seq_len_multiple_of = seq_len_multiple_of
        self.output_layer = output_layer

        if device is not None:
            self.to(device)

        model_path = Path(checkpoint_path)

        assert model_path.exists(), f'path {checkpoint_path} does not exist'

        checkpoint = torch.load(checkpoint_path, map_location=device)
        load_model_input = {checkpoint_path: checkpoint}
        model, *_ = fairseq.checkpoint_utils.load_model_ensemble_and_task(load_model_input)

        if device is not None:
            model[0].to(device)

        self.model = model[0]
        self.model.eval()

    @property
    def groups(self):
        return 1

    @torch.no_grad()
    def forward(
        self,
        wav_input,
        flatten=True,
        input_sample_hz=None
    ):
        device = wav_input.device

        if exists(input_sample_hz):
            wav_input = resample(wav_input, input_sample_hz, self.target_sample_hz)

        #if exists(self.seq_len_multiple_of):
            #wav_input = curtail_to_multiple(wav_input, self.seq_len_multiple_of)

        embed = self.model(
            wav_input,
            features_only=True,
            mask=False,  # thanks to @maitycyrus for noticing that mask is defaulted to True in the fairseq code
            output_layer=self.output_layer
        )

        embed, packed_shape = pack([embed['x']], '* d')

        # codebook_indices = self.kmeans.predict(embed.cpu().detach().numpy())

        codebook_indices = torch.from_numpy(embed.cpu().detach().numpy()).to(device)  # .long()

        if flatten:
            return codebook_indices

        codebook_indices, = unpack(codebook_indices, packed_shape, '*')
        return codebook_indices



import re
import unicodedata


def encode_path_safe(filename: str, allow_unicode=False):
    """
    Makes a string path safe by removing / replacing not by the os allowed patterns.
    This converts:
    spaces 2 dashes, repeated dashes 2 single dashes, remove non alphanumerics, underscores, or hyphen, string 2 lowercase
    strip
    """
    filename = str(filename)
    if allow_unicode:
        filename = unicodedata.normalize('NFKC', filename)
    else:
        filename = unicodedata.normalize('NFKD', filename).encode('ascii', 'ignore').decode('ascii')
    filename = re.sub(r'[^\w\s-]', '', filename.lower())
    return re.sub(r'[-\s]+', '', filename).strip('-_')


def get_cpu_or_gpu() -> str:
    import torch
    if torch.cuda.is_available():
        return 'cuda'
    else:
        return 'cpu'
